@@ -89,7 +89,7 @@ static int smtp_code (char *buf, size_t len, int *n)
   code[1] = buf[1];
   code[2] = buf[2];
   code[3] = 0;
-  if (mutt_atoi (code, n) < 0)
+  if (mutt_atoi (code, n, 0) < 0)
     return -1;
   return 0;
 }
@@ -489,17 +489,8 @@ static int smtp_open (CONNECTION* conn)
   }
 #endif
 
-  if (conn->account.flags & MUTT_ACCT_USER)
-  {
-    if (!mutt_bit_isset (Capabilities, AUTH))
-    {
-      mutt_error (_("SMTP server does not support authentication"));
-      mutt_sleep (1);
-      return -1;
-    }
-
+  if (mutt_bit_isset (Capabilities, AUTH))
     return smtp_auth (conn);
-  }
 
   return 0;
 }
